"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/* tslint:disable:object-literal-sort-keys */
// tslint:disable:no-var-requires
require("whatwg-fetch");
var fetchEngine_1 = require("../fetchEngine");
var FetchGroup_1 = require("../FetchGroup");
function mockFetch(request) {
    return Promise.resolve(new Response());
}
it("makeFetchEngine is requireable", function () {
    expect(fetchEngine_1.default).toBeDefined();
});
it("fetchEngine with no args returns a response", function () {
    expect.assertions(1);
    var fetch = fetchEngine_1.default(mockFetch)();
    return fetch(new Request("/mock"))
        .then(function (res) {
        expect(res).toBeDefined();
    });
});
it("fetchEngine with shouldFetch false throws", function () {
    expect.assertions(1);
    var fetch = fetchEngine_1.default(mockFetch)(new FetchGroup_1.default({
        plugins: [
            {
                shouldFetch: function () { return false; },
            },
        ],
    }));
    return expect(fetch(new Request("/mock"))).rejects.toThrow("shouldFetch prevented the request from being made");
});
it("fetchEngine with getRequest passes transformed request to willFetch", function () {
    expect.assertions(2);
    var firstMockReq = new Request("/mock");
    var secondMockReq = new Request("/mock/other");
    var fetch = fetchEngine_1.default(mockFetch)(new FetchGroup_1.default({
        plugins: [
            {
                getRequest: function (req) {
                    expect(req).toEqual(firstMockReq);
                    return Promise.resolve(secondMockReq);
                },
                willFetch: function (req) {
                    expect(req).toEqual(secondMockReq);
                },
            },
        ],
    }));
    return fetch(firstMockReq);
});
it("fetchEngine with fetch passes request and next function", function () {
    expect.assertions(2);
    var mockReq = new Request("/mock");
    var fetch = fetchEngine_1.default(mockFetch)(new FetchGroup_1.default({
        plugins: [
            {
                fetch: function (req, next) {
                    expect(req).toEqual(mockReq);
                    expect(typeof next).toBe("function");
                    return next();
                },
            },
        ],
    }));
    return fetch(mockReq);
});
it("fetchEngine with fetching passes promise and original request", function () {
    expect.assertions(3);
    var mockReq = new Request("/mock");
    var fetch = fetchEngine_1.default(mockFetch)(new FetchGroup_1.default({
        plugins: [
            {
                fetching: function (args) {
                    expect(args.request).toEqual(mockReq);
                    expect(args.promise).toBeDefined();
                    expect(typeof args.promise.then).toBe("function");
                },
            },
        ],
    }));
    return fetch(mockReq);
});
it("fetchEngine with getResponse passes transformed response to didFetch", function () {
    expect.assertions(1);
    var mockReq = new Request("/mock");
    var mockRes = new Response();
    var fetch = fetchEngine_1.default(mockFetch)(new FetchGroup_1.default({
        plugins: [
            {
                getResponse: function (res) {
                    return Promise.resolve(mockRes);
                },
                didFetch: function (res) {
                    expect(res).toBe(mockRes);
                },
            },
        ],
    }));
    return fetch(mockReq);
});
it("fetchEngine with getResponse passes root fetch method for retries", function () {
    expect.assertions(2);
    var mockReq = new Request("/mock");
    var mockRes = new Response();
    var fetch = fetchEngine_1.default(mockFetch)(new FetchGroup_1.default({
        plugins: [
            {
                getResponse: function (res, retry) {
                    expect(typeof retry).toBe("function");
                    return Promise.resolve(mockRes);
                },
            },
        ],
    }));
    return fetch(mockReq).then(function (res) {
        expect(res).toBe(mockRes);
    });
});
it("fetchEngine with getResponse can retry original request", function () {
    expect.assertions(2);
    var calls = 0;
    var mockReq = new Request("/mock");
    var mockRes = new Response();
    var fetch = fetchEngine_1.default(mockFetch)(new FetchGroup_1.default({
        plugins: [
            {
                willFetch: function () {
                    calls = calls + 1;
                },
                getResponse: function (res, retry) {
                    if (calls === 1) {
                        return retry();
                    }
                    else {
                        return mockRes;
                    }
                },
            },
        ],
    }));
    return fetch(mockReq).then(function (res) {
        expect(calls).toBe(2);
        expect(res).toBe(mockRes);
    });
});
it("fetchEngine with getResponse can retry new request", function () {
    expect.assertions(4);
    var calls = 0;
    var mockReq = new Request("/mock");
    var mockRetryReq = new Request("/mock-retry");
    var mockRes = new Response();
    var fetch = fetchEngine_1.default(mockFetch)(new FetchGroup_1.default({
        plugins: [
            {
                willFetch: function (req) {
                    calls = calls + 1;
                    if (calls === 1) {
                        expect(req).toEqual(mockReq);
                    }
                    else {
                        expect(req).toEqual(mockRetryReq);
                    }
                },
                getResponse: function (res, retry) {
                    if (calls === 1) {
                        return retry(mockRetryReq);
                    }
                    else {
                        return mockRes;
                    }
                },
            },
        ],
    }));
    return fetch(mockReq).then(function (res) {
        expect(calls).toBe(2);
        expect(res).toEqual(mockRes);
    });
});
it("fetchEngine flows through full stack in order", function () {
    expect.assertions(6);
    var firstMockReq = new Request("/mock");
    var secondMockReq = new Request("/mock");
    var firstMockRes = new Response();
    var fetch = fetchEngine_1.default(mockFetch)(new FetchGroup_1.default({
        plugins: [
            {
                shouldFetch: function (req) {
                    expect(req).toEqual(firstMockReq);
                    return true;
                },
                getRequest: function (req) {
                    expect(req).toEqual(firstMockReq);
                    return Promise.resolve(secondMockReq);
                },
                willFetch: function (req) {
                    expect(req).toEqual(secondMockReq);
                },
                fetch: function (req, next) {
                    expect(req).toEqual(secondMockReq);
                    return next();
                },
                fetching: function (args) {
                    expect(args.request).toEqual(secondMockReq);
                },
                getResponse: function (res, retry) {
                    // TODO find a way to fake the retured response
                    return Promise.resolve(firstMockRes);
                },
                didFetch: function (res) {
                    expect(res).toEqual(firstMockRes);
                },
            },
        ],
    }));
    return fetch(firstMockReq);
});
it("fetchEngine with opts creates FetchGroup and flows through full stack in order", function () {
    expect.assertions(12);
    var firstMockReq = new Request("/mock");
    var secondMockReq = new Request("/mock");
    var firstMockRes = new Response();
    var fetch = fetchEngine_1.default(mockFetch)({
        filters: [
            {
                testRequest: function (req) {
                    expect(req).toEqual(firstMockReq);
                    return true;
                },
                testResponse: function (res) {
                    expect(res).toEqual(firstMockRes);
                    return true;
                },
            },
        ],
        plugins: [
            {
                shouldFetch: function (req) {
                    expect(req).toEqual(firstMockReq);
                    return true;
                },
                getRequest: function (req) {
                    expect(req).toEqual(firstMockReq);
                    return Promise.resolve(secondMockReq);
                },
                willFetch: function (req) {
                    expect(req).toEqual(secondMockReq);
                },
                fetch: function (req, next) {
                    expect(req).toEqual(secondMockReq);
                    return next();
                },
                fetching: function (args) {
                    expect(args.request).toEqual(secondMockReq);
                },
                getResponse: function (res) {
                    // TODO find a way to fake the retured response
                    return Promise.resolve(firstMockRes);
                },
                didFetch: function (res) {
                    expect(res).toEqual(firstMockRes);
                },
            },
        ],
    });
    return fetch(firstMockReq);
});
//# sourceMappingURL=fetchEngine.test.js.map