"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var FetchGroup_1 = require("./FetchGroup");
var sideEffect_1 = require("./utils/sideEffect");
// makeFetchEngine takes an implementation of fetch so that we can make
// versions of the library for node and the browser.
function makeFetchEngine(innerFetch) {
    // Here is the actual implementation of the fetch-engine, where we've
    // closed over the 'inner' fetch implementation.
    return function fetchEngine(pluginOrOpts) {
        if (pluginOrOpts === void 0) { pluginOrOpts = new FetchGroup_1.default(); }
        // Marshal the input to make sure it's a valid plugin that implements all the methods
        // that we assume to be in place below
        var plugin = (isPlugin(pluginOrOpts) ?
            pluginOrOpts :
            new FetchGroup_1.default(pluginOrOpts));
        return function fetch(input, init) {
            // Normalise the fetch API sugar into a Request
            var originalRequest = new Request(input, init);
            // In order to retry we have to close over the original input
            var retry = function (req) {
                if (typeof req !== "undefined") {
                    return fetch(req);
                }
                else {
                    return fetch(originalRequest);
                }
            };
            var pReq = Promise.resolve(originalRequest);
            var pShouldFetch = pReq.then(sideEffect_1.default(function (request) {
                if (!plugin.shouldFetch(request)) {
                    throw new Error("shouldFetch prevented the request from being made");
                }
            }));
            var pGetRequest = pShouldFetch.then(plugin.getRequest.bind(plugin));
            var pWillFetch = pGetRequest.then(sideEffect_1.default(plugin.willFetch.bind(plugin)));
            var pFetching = pWillFetch.then(function (request) {
                return Promise.resolve(plugin.fetch(request, function () {
                    // Business time.
                    var pFetch = innerFetch(request);
                    // Side effects!
                    plugin.fetching({
                        promise: pFetch,
                        request: request,
                    });
                    return pFetch;
                }));
            });
            var pGetResponse = pFetching.then(function (res) {
                return plugin.getResponse(res, retry);
            });
            var pDidFetch = pGetResponse.then(sideEffect_1.default(plugin.didFetch.bind(plugin)));
            return pDidFetch;
        };
    };
}
exports.default = makeFetchEngine;
function isPlugin(o) {
    if (o === void 0) { o = {}; }
    return [
        "shouldFetch", "getRequest", "willFetch",
        "fetch", "fetching",
        "getResponse", "didFetch",
    ].every(function (k) { return typeof o[k] === "function"; });
}
//# sourceMappingURL=fetchEngine.js.map