"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Compose an array of functions that take two arguments: a value of type T,
 * and a "next" function that passes control onto the next function in the
 * chain, finally reaching a "base" function (supplied at call time).
 *
 * An example function:
 *
 *    var lessThanThree = (v, next) => {
 *      return (v < 3 ? true : next());
 *    };
 *
 * Example usage:
 *
 *    var f = composeContinuation([lessThanThree, greaterThanSeven]);
 *    f(5, () => false);
 */
function composeContinuation(fns) {
    return fns.reduceRight(
    // 'next' is always a function that takes a value (some T) and a 'base'
    // function to call if needs be. The calling of that function is taken care
    // of by the reduce becuase it is passed all the way through the chain.
    // Each 'fn' is a input function, in that it does not need to pass its
    // value argument (some T) through to the 'next' function — it can just
    // 'return next()', and rely on the original argument to be passed through.
    // This also prevents users altering the input argument (apart from in-place
    // mutation).
    function (next, fn) {
        // We return another composed function, which will be the 'next' for the
        // next input function.
        return function (v, base) {
            // The innermost function here tries each input function in turn, but
            // allows it to pass control onto the next in the chain.
            return fn(v, function () { return next(v, base); });
        };
    }, 
    // This is the innermost function of the composition. It just passes the
    // value (some T) into the base function that will have been passed through
    // the chain up to this point.
    function (v, base) { return base(v); });
}
exports.default = composeContinuation;
//# sourceMappingURL=composeContinuation.js.map